#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020

import copy
import datetime
import json
import random
import os
import traceback

from idds.common import exceptions
from idds.common.constants import (TransformType, CollectionType, CollectionStatus,
                                   ContentStatus, ContentType,
                                   ProcessingStatus, WorkStatus)
from idds.common.utils import run_command
from idds.common.utils import replace_parameters_with_values
# from idds.workflow.work import Work
from idds.workflow.work import Processing
from idds.atlas.workflow.atlascondorwork import ATLASCondorWork
# from idds.core import (catalog as core_catalog)


class ATLASHPOWork(ATLASCondorWork):
    def __init__(self, executable=None, arguments=None, parameters=None, setup=None,
                 work_tag='hpo', exec_type='local', sandbox=None, work_id=None,
                 name=None,
                 # primary_input_collection=None, other_input_collections=None,
                 # output_collections=None, log_collections=None,
                 logger=None,
                 workload_id=None,
                 agent_attributes=None,
                 method=None,
                 container_workdir=None,
                 output_json=None,
                 opt_space=None, initial_points=None,
                 max_points=None, num_points_per_iteration=10):
        """
        Init a work/task/transformation.

        :param setup: A string to setup the executable enviroment, it can be None.
        :param executable: The executable.
        :param arguments: The arguments.
        :param parameters: A dict with arguments needed to be replaced.
        :param work_type: The work type like data carousel, hyperparameteroptimization and so on.
        :param exec_type: The exec type like 'local', 'remote'(with remote_package set), 'docker' and so on.
        :param sandbox: The sandbox.
        :param work_id: The work/task id.
        :param primary_input_collection: The primary input collection.
        :param other_input_collections: List of the input collections.
        :param output_collections: List of the output collections.
        # :param workflow: The workflow the current work belongs to.
        :param method: The HPO methd to use. It can be 'nevergrad', 'container' or 'sandbox'.
        :param sandbox: The sandbox to be uploaded or the container path.
        :param executable: The executable command.
        :param arguments: The arguments for the executable.
        :param container_workdir: The working directory for container.
        :param opt_space: The optimization space.
        :param initial_points: The initial points.
        :param max_points: The maximum number of points.
        :param number_points_per_iteration: The number of points to be generated per iteration.
        """
        if not name:
            if workload_id:
                name = 'hpo.' + str(workload_id) + "." + datetime.datetime.utcnow().strftime("%Y_%m_%d_%H_%M_%S_%f") + str(random.randint(1, 1000))
            else:
                name = 'hpo.' + datetime.datetime.utcnow().strftime("%Y_%m_%d_%H_%M_%S_%f") + str(random.randint(1, 1000))

        primary_input_collection = {'scope': 'HPO', 'name': name}
        other_input_collections = None
        output_collections = [{'scope': 'HPO', 'name': name + 'output'}]
        log_collections = None

        super(ATLASHPOWork, self).__init__(executable=executable, arguments=arguments,
                                           parameters=parameters, setup=setup, work_type=TransformType.HyperParameterOpt,
                                           exec_type=exec_type, sandbox=sandbox, work_id=work_id,
                                           primary_input_collection=primary_input_collection,
                                           other_input_collections=other_input_collections,
                                           output_collections=output_collections,
                                           log_collections=log_collections,
                                           logger=logger,
                                           agent_attributes=agent_attributes)
        self.method = method
        self.sandbox = sandbox
        self.executable = executable
        self.arguments = arguments
        self.container_workdir = container_workdir
        self.opt_space = opt_space
        self.initial_points = initial_points
        self.max_points = max_points
        self.num_points_per_iteration = num_points_per_iteration
        self.unfinished_points = 0

        self.input_json = None
        self.output_json = output_json

        self.finished_points = 0
        self.points_to_generate = self.num_points_per_iteration
        self.point_index = 0
        self.terminated = False
        self.polling_retries = 0

        if not self.num_points_per_iteration or self.num_points_per_iteration < 0:
            raise exceptions.IDDSException("num_points_per_iteration must be integer bigger than 0")
        self.num_points_per_iteration = int(self.num_points_per_iteration)

        if not self.method and self.executable and 'docker' in self.executable:
            self.method = 'docker'

        # if self.agent_attributes and 'atlashpowork' in self.agent_attributes:
        #    self.agent_attributes = self.agent_attributes['atlashpowork']
        # self.logger.info("agent_attributes: %s" % self.agent_attributes)

        # if self.agent_attributes and 'workdir' in self.agent_attributes and self.agent_attributes['workdir']:
        #     self.set_workdir(self.agent_attributes['workdir'])
        # self.logger.info("workdir: %s" % self.get_workdir())

        if agent_attributes:
            self.set_agent_attributes(agent_attributes)

    def set_agent_attributes(self, attrs, req_attributes=None):
        self.agent_attributes = attrs

        if self.agent_attributes and 'atlashpowork' in self.agent_attributes:
            self.agent_attributes = self.agent_attributes['atlashpowork']
        self.logger.info("agent_attributes: %s" % self.agent_attributes)

        if self.agent_attributes and 'workdir' in self.agent_attributes and self.agent_attributes['workdir']:
            if req_attributes and 'request_id' in req_attributes and 'workload_id' in req_attributes and 'transform_id' in req_attributes:
                req_dir = 'request_%s_%s/transform_%s' % (req_attributes['request_id'],
                                                          req_attributes['workload_id'],
                                                          req_attributes['transform_id'])
                self.set_workdir(os.path.join(self.agent_attributes['workdir'], req_dir))
        self.logger.info("workdir: %s" % self.get_workdir())

    ####### functions for transformer ########   # noqa E266
    ######################################       # noqa E266

    def poll_external_collection(self, coll):
        try:
            if coll.status in [CollectionStatus.Closed]:
                if not self.terminated:
                    self.logger.info("Work is not terminated, reopen collection")
                    coll.coll_metadata['is_open'] = True
                    coll.status = CollectionStatus.Open
                return coll
            else:
                coll.coll_metadata['bytes'] = 0
                coll.coll_metadata['total_files'] = 0
                coll.coll_metadata['availability'] = True
                coll.coll_metadata['events'] = 0
                coll.coll_metadata['is_open'] = True
                coll.coll_metadata['run_number'] = None
                coll.coll_metadata['did_type'] = 'DATASET'
                coll.coll_metadata['list_all_files'] = False

                if self.terminated:
                    self.logger.info("Work is terminated. Closing input dataset.")
                    coll.coll_metadata['is_open'] = False

                if self.points_to_generate <= 0:
                    self.logger.info("points_to_generate(%s) is equal or smaller than 0. Closing input dataset." % self.points_to_generate)
                    coll.coll_metadata['is_open'] = False

                if 'is_open' in coll.coll_metadata and not coll.coll_metadata['is_open']:
                    coll_status = CollectionStatus.Closed
                else:
                    coll_status = CollectionStatus.Open
                coll.status = coll_status

                if 'did_type' in coll.coll_metadata:
                    if coll.coll_metadata['did_type'] == 'DATASET':
                        coll_type = CollectionType.Dataset
                    elif coll.coll_metadata['did_type'] == 'CONTAINER':
                        coll_type = CollectionType.Container
                    else:
                        coll_type = CollectionType.File
                else:
                    coll_type = CollectionType.Dataset
                coll.coll_metadata['coll_type'] = coll_type

                return coll
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))

    def get_input_collections(self):
        # return [self.primary_input_collection] + self.other_input_collections
        colls = [self.primary_input_collection] + self.other_input_collections
        for coll_int_id in colls:
            coll = self.collections[coll_int_id]
            coll = self.poll_external_collection(coll)
            self.collections[coll_int_id] = coll
        return super(ATLASHPOWork, self).get_input_collections()

    def get_input_contents(self, point_index=1):
        """
        Get all input contents from DDM.
        """
        try:
            if self.terminated:
                return []

            if self.unfinished_points > 0:
                return []

            ret_files = []
            coll = self.collections[self.primary_input_collection]

            if self.max_points and (self.max_points - self.finished_points < self.num_points_per_iteration):
                self.points_to_generate = self.max_points - self.finished_points

            # call external processing to generate points
            points = self.generate_points()
            self.logger.info("points generated: %s" % str(points))

            loss = None
            for point in points:
                ret_file = {'coll_id': coll.coll_id,
                            'scope': coll.scope,
                            'name': str(point_index),
                            'bytes': 0,
                            'adler32': None,
                            'min_id': 0,
                            'max_id': 0,
                            'path': json.dumps((point, loss)),
                            'content_type': ContentType.File,
                            'content_metadata': {'events': 0}}
                ret_files.append(ret_file)
                point_index += 1
            return ret_files
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))

    def get_mapped_inputs(self, mapped_input_output_maps):
        ret = []
        for map_id in mapped_input_output_maps:
            inputs = mapped_input_output_maps[map_id]['inputs']

            # if 'primary' is not set, the first one is the primary input.
            primary_input = inputs[0]
            for ip in inputs:
                if 'primary' in ip['content_metadata'] and ip['content_metadata']['primary']:
                    primary_input = ip
            ret.append(primary_input)
        return ret

    def get_unfinished_points(self, mapped_input_output_maps):
        counts = 0
        count_finished = 0
        for map_id in mapped_input_output_maps:
            outputs = mapped_input_output_maps[map_id]['outputs']

            for op in outputs:
                if op['status'] in [ContentStatus.New]:
                    counts += 1
                if op['status'] in [ContentStatus.Available]:
                    count_finished += 1
        self.finished_points = count_finished
        return counts

    def get_new_input_output_maps(self, mapped_input_output_maps={}):
        """
        New inputs which are not yet mapped to outputs.

        :param mapped_input_output_maps: Inputs that are already mapped.
        """
        unfinished_mapped = self.get_unfinished_points(mapped_input_output_maps)
        self.unfinished_points = unfinished_mapped

        mapped_inputs = self.get_mapped_inputs(mapped_input_output_maps)
        mapped_inputs_scope_name = [ip['scope'] + ":" + ip['name'] for ip in mapped_inputs]
        mapped_keys = mapped_input_output_maps.keys()
        if mapped_keys:
            next_key = max(mapped_keys) + 1
        else:
            next_key = 1

        inputs = self.get_input_contents(point_index=next_key)

        new_inputs = []
        new_input_output_maps = {}
        for ip in inputs:
            ip_scope_name = ip['scope'] + ":" + ip['name']
            if ip_scope_name not in mapped_inputs_scope_name:
                new_inputs.append(ip)

        # to avoid cheking new inputs if there are no new inputs anymore
        if (not new_inputs and self.collections[self.primary_input_collection]
           and self.collections[self.primary_input_collection].status in [CollectionStatus.Closed]):  # noqa: W503
            self.set_has_new_inputs(False)
        else:
            for ip in new_inputs:
                out_ip = copy.deepcopy(ip)
                ip['status'] = ContentStatus.Available
                ip['substatus'] = ContentStatus.Available
                out_ip['coll_id'] = self.collections[self.output_collections[0]].coll_id
                new_input_output_maps[next_key] = {'inputs': [ip],
                                                   'outputs': [out_ip],
                                                   'inputs_dependency': [],
                                                   'logs': []}
                next_key += 1

        self.unfinished_points = self.unfinished_points + len(new_inputs)

        return new_input_output_maps

    def generate_points(self):
        active_processing = self.get_processing(None, without_creating=True)
        if not active_processing:
            if self.points_to_generate > 0:
                active_processing = self.create_processing(None)
                log_str = "max_points: %s, finished_points: %s, points_to_generate: %s, new processing: %s" % (self.max_points,
                                                                                                               self.finished_points,
                                                                                                               self.points_to_generate,
                                                                                                               active_processing)
                self.logger.info(log_str)
                if active_processing:
                    return []
                else:
                    self.polling_retries += 1
                    if self.polling_retries > 3:
                        self.terminated = True
                        self.set_terminated_msg("Failed to create processing")
                    return []
            else:
                self.polling_retries += 1
                if self.polling_retries > 3:
                    self.terminated = True
                self.set_terminated_msg("Number of points is enough(points_to_generate: %s)" % self.points_to_generate)
                return []

        if active_processing and self.is_processing_terminated(active_processing):
            self.logger.info("processing terminated: %s" % active_processing)
            self.reap_processing(active_processing)
            # output_metadata = active_processing['output_metadata']
            output_metadata = active_processing.output_data
            if output_metadata:
                self.polling_retries = 0
                if self.max_points and self.max_points > self.finished_points:
                    self.terminated = False
                return output_metadata
            else:
                self.polling_retries += 1
                if self.polling_retries > 3:
                    self.terminated = True
                # processing_metadata = active_processing['processing_metadata']
                # errors = None
                # if 'errors' in processing_metadata:
                #     errors = processing_metadata['errors']
                errors = active_processing.errors
                self.set_terminated_msg("No points generated. Terminating the Work/Transformation. Detailed errors: %s" % errors)
                return []
        return []

    def get_processing(self, input_output_maps, without_creating=False):
        if self.active_processings:
            return self.processings[self.active_processings[0]]
        else:
            # if not without_creating:
            #     return self.create_processing(input_output_maps)
            pass
        return None

    def create_processing(self, input_output_maps=[]):
        processing_metadata = {'points_to_generate': self.points_to_generate}
        proc = Processing(processing_metadata=processing_metadata)
        self.add_processing_to_processings(proc)
        self.active_processings.append(proc.internal_id)
        return proc

    def get_status_statistics(self, registered_input_output_maps):
        status_statistics = {}
        for map_id in registered_input_output_maps:
            outputs = registered_input_output_maps[map_id]['outputs']

            for content in outputs:
                if content['status'].name not in status_statistics:
                    status_statistics[content['status'].name] = 0
                status_statistics[content['status'].name] += 1
        self.status_statistics = status_statistics
        return status_statistics

    """
    def syn_collection_status(self):
        input_collections = self.get_input_collections()
        output_collections = self.get_output_collections()
        # log_collections = self.get_log_collections()

        for input_collection in input_collections:
            input_collection['total_files'] = self.finished_points + self.unfinished_points
            input_collection['processed_files'] = self.finished_points + self.unfinished_points

        for output_collection in output_collections:
            output_collection['total_files'] = self.finished_points + self.unfinished_points
            output_collection['processed_files'] = self.finished_points
    """

    def syn_work_status(self, registered_input_output_maps, all_updates_flushed=True, output_statistics={}, to_release_input_contents=[]):
        super(ATLASHPOWork, self).syn_work_status(registered_input_output_maps)
        self.get_status_statistics(registered_input_output_maps)

        # self.syn_collection_status()

        if self.is_processings_terminated() and not self.has_new_inputs:
            if not self.is_all_outputs_flushed(registered_input_output_maps):
                self.logger.warn("The processing is terminated. but not all outputs are flushed. Wait to flush the outputs then finish the transform")
                return

            keys = self.status_statistics.keys()
            if len(keys) == 1:
                if ContentStatus.Available.name in keys:
                    self.status = WorkStatus.Finished
                else:
                    self.status = WorkStatus.Failed
            else:
                self.status = WorkStatus.SubFinished
        else:
            self.status = WorkStatus.Transforming

    ####### functions for carrier ########     # noqa E266
    ######################################     # noqa E266

    def generate_processing_script_nevergrad(self, processing):
        executable = self.agent_attributes['nevergrad']['executable']
        arguments = self.agent_attributes['nevergrad']['arguments']

        param_values = {'MAX_POINTS': self.max_points,
                        'NUM_POINTS': self.points_to_generate,
                        'IN': self.input_json,
                        'OUT': self.output_json}
        if 'X509_USER_PROXY' in os.environ and os.environ['X509_USER_PROXY']:
            proxy_filename = os.path.basename(os.environ['X509_USER_PROXY'])
            param_values['X509_USER_PROXY_FULLNAME'] = os.environ['X509_USER_PROXY']
            param_values['X509_USER_PROXY_BASENAME'] = proxy_filename

        arguments = replace_parameters_with_values(arguments, param_values)

        script = "#!/bin/bash\n\n"
        script += "executable=%s\n" % os.path.basename(executable)
        script += "arguments='%s'\n" % str(arguments)
        script += "input_json=%s\n" % str(self.input_json)
        script += "output_json=%s\n" % str(self.output_json)
        script += "\n"

        script += "env\n"
        script += "echo $X509_USER_PROXY\n"
        script += "\n"

        script += "echo 'user id:'\n"
        script += "id\n"
        script += "\n"

        script += "echo '%s' '%s'\n" % (os.path.basename(executable), str(arguments))
        script += '%s %s\n' % (os.path.basename(executable), str(arguments))

        script += '\n'

        long_id = self.get_long_id(processing)
        script_name = 'processing_%s.sh' % long_id
        script_name = os.path.join(self.get_working_dir(processing), script_name)
        with open(script_name, 'w') as f:
            f.write(script)
        run_command("chmod +x %s" % script_name)
        return script_name

    def generate_processing_script_container(self, processing):
        param_values = {'MAX_POINTS': self.max_points,
                        'NUM_POINTS': self.points_to_generate,
                        'IN': self.input_json,
                        'OUT': self.output_json}
        proxy_filename = 'x509up'
        if 'X509_USER_PROXY' in os.environ and os.environ['X509_USER_PROXY']:
            proxy_filename = os.path.basename(os.environ['X509_USER_PROXY'])
            param_values['X509_USER_PROXY_FULLNAME'] = os.environ['X509_USER_PROXY']
            param_values['X509_USER_PROXY_BASENAME'] = proxy_filename

        executable = replace_parameters_with_values(self.executable, param_values)
        arguments = replace_parameters_with_values(self.arguments, param_values)

        script = "#!/bin/bash\n\n"
        script += "executable=%s\n" % str(executable)
        script += "arguments=%s\n" % str(arguments)
        script += "input_json=%s\n" % str(self.input_json)
        script += "output_json=%s\n" % str(self.output_json)
        script += "\n"

        script += "env\n"
        script += "echo $X509_USER_PROXY\n"
        script += "\n"

        script += "echo 'user id:'\n"
        script += "id\n"
        script += "\n"

        if self.sandbox and 'docker' in executable:
            arguments = 'run --rm -v $(pwd):%s -v /cvmfs:/cvmfs -e X509_USER_PROXY=%s/%s %s ' % (self.container_workdir, self.container_workdir, proxy_filename, self.sandbox) + arguments

        script += "echo '%s' '%s'\n" % (str(executable), str(arguments))
        script += '%s %s\n' % (str(executable), str(arguments))

        if self.sandbox and 'docker' in executable:
            script += 'docker image rm -f %s\n' % self.sandbox

        script += '\n'

        long_id = self.get_long_id(processing)
        script_name = 'processing_%s.sh' % long_id
        script_name = os.path.join(self.get_working_dir(processing), script_name)
        with open(script_name, 'w') as f:
            f.write(script)
        run_command("chmod +x %s" % script_name)
        return script_name

    def generate_processing_script_sandbox(self, processing):
        param_values = {'MAX_POINTS': self.max_points,
                        'NUM_POINTS': self.points_to_generate,
                        'IN': self.input_json,
                        'OUT': self.output_json}
        if 'X509_USER_PROXY' in os.environ and os.environ['X509_USER_PROXY']:
            proxy_filename = os.path.basename(os.environ['X509_USER_PROXY'])
            param_values['X509_USER_PROXY_FULLNAME'] = os.environ['X509_USER_PROXY']
            param_values['X509_USER_PROXY_BASENAME'] = proxy_filename

        executable = replace_parameters_with_values(self.executable, param_values)
        arguments = replace_parameters_with_values(self.arguments, param_values)

        script = "#!/bin/bash\n\n"
        script += "sandbox=%s\n" % str(self.sandbox)
        script += "executable=%s\n" % str(executable)
        script += "arguments=%s\n" % str(arguments)
        script += "input_json=%s\n" % str(self.input_json)
        script += "output_json=%s\n" % str(self.output_json)
        script += "\n"

        script += "env\n"
        script += "echo $X509_USER_PROXY\n"
        script += "\n"

        script += "echo 'user id:'\n"
        script += "id\n"
        script += "\n"

        script += "wget $sandbox\n"
        script += 'base_sandbox="$(basename -- $sandbox)"\n'
        script += 'tar xzf $base_sandbox\n'

        script += 'chmod +x %s\n' % str(executable)
        script += "echo '%s' '%s'\n" % (str(executable), str(arguments))
        script += '%s %s\n' % (str(executable), str(arguments))

        script += '\n'

        long_id = self.get_long_id(processing)
        script_name = 'processing_%s.sh' % long_id
        script_name = os.path.join(self.get_working_dir(processing), script_name)
        with open(script_name, 'w') as f:
            f.write(script)
        run_command("chmod +x %s" % script_name)
        return script_name

    def generate_input_json(self, processing):
        try:
            from idds.core import (catalog as core_catalog)

            output_collection = self.get_output_collections()[0]
            contents = core_catalog.get_contents_by_coll_id_status(coll_id=output_collection.coll_id)
            points = []
            for content in contents:
                # point = content['content_metadata']['point']
                point = json.loads(content['path'])
                points.append(point)

            job_dir = self.get_working_dir(processing)
            if 'input_json' in self.agent_attributes and self.agent_attributes['input_json']:
                input_json = self.agent_attributes['input_json']
            else:
                input_json = 'idds_input.json'
            opt_points = {'points': points, 'opt_space': self.opt_space}
            with open(os.path.join(job_dir, input_json), 'w') as f:
                json.dump(opt_points, f)
            return input_json
        except Exception as e:
            raise Exception("Failed to generate idds inputs for HPO: %s" % str(e))

    def get_output_json(self, processing):
        # job_dir = self.get_working_dir(processing)
        if self.output_json:
            return self.output_json
        elif 'output_json' in self.agent_attributes and self.agent_attributes['output_json']:
            output_json = self.agent_attributes['output_json']
        else:
            output_json = 'idds_output.json'
        return output_json

    def generate_processing_script(self, processing):
        if not self.method:
            err_msg = "Processing %s HPO method(%s) is not defined" % (processing['processing_id'], self.method)
            self.logger.error(err_msg)
            self.set_terminated_msg(err_msg)
            self.terminated = True
            return None, err_msg

        self.input_json = self.generate_input_json(processing)
        self.output_json = self.get_output_json(processing)

        if self.method == 'nevergrad':
            script_name = self.generate_processing_script_nevergrad(processing)
            return script_name, None
        elif self.method in ['container', 'docker']:
            script_name = self.generate_processing_script_container(processing)
            return script_name, None
        elif self.method == 'sandbox':
            script_name = self.generate_processing_script_sandbox(processing)
            return script_name, None
        else:
            err_msg = "Processing %s not supported HPO method: %s" % (processing['processing_id'], self.method)
            self.logger.error(err_msg)
            self.set_terminated_msg(err_msg)
            self.terminated = True
            return None, err_msg

    def get_input_files(self, processing):
        return [self.input_json]

    def get_output_files(self, processing):
        return [self.output_json]

    def submit_processing(self, processing):
        proc = processing['processing_metadata']['processing']
        if proc.external_id:
            # if 'job_id' in processing['processing_metadata']:
            pass
        else:
            job_id, errors = self.submit_condor_processing(processing)
            # processing['processing_metadata']['job_id'] = job_id
            # processing['processing_metadata']['errors'] = errors
            proc.external_id = job_id
            if job_id:
                proc.submitted_at = datetime.datetime.utcnow()
            proc.errors = errors

    def parse_processing_outputs(self, processing):
        request_id = processing['request_id']
        workload_id = processing['workload_id']
        processing_id = processing['processing_id']

        if not self.output_json:
            return None, 'Request(%s)_workload(%s)_processing(%s) output_json(%s) is not defined' % (request_id, workload_id,
                                                                                                     processing_id, self.output_json)

        job_dir = self.get_working_dir(processing)
        full_output_json = os.path.join(job_dir, self.output_json)
        if not os.path.exists(full_output_json):
            return None, '%s is not created' % str(full_output_json)
        else:
            try:
                with open(full_output_json, 'r') as f:
                    data = f.read()
                outputs = json.loads(data)
                if not outputs:
                    return outputs, "No points generated: the outputs is empty"
                return outputs, None
            except Exception as ex:
                return None, 'Failed to load the content of %s: %s' % (str(full_output_json), str(ex))

    def poll_processing(self, processing):
        try:
            proc = processing['processing_metadata']['processing']
            job_status, job_err_msg = self.poll_condor_job_status(processing, proc.external_id)
            processing_outputs = None
            reset_expired_at = False
            if job_status in [ProcessingStatus.Finished]:
                job_outputs, parser_errors = self.parse_processing_outputs(processing)
                if job_outputs:
                    processing_status = ProcessingStatus.Finished
                    processing_err = None
                    processing_outputs = job_outputs
                else:
                    processing_status = ProcessingStatus.Failed
                    processing_err = parser_errors
            elif job_status in [ProcessingStatus.Failed]:
                processing_status = job_status
                processing_err = job_err_msg
            elif self.toexpire:
                processing_status = ProcessingStatus.Expired
                processing_err = "The processing is expired"
            elif job_status in [ProcessingStatus.Cancelled]:
                processing_status = job_status
                processing_err = job_err_msg
            elif self.tocancel:
                self.cancelled_processings.append(proc.internal_id)
                processing_status = ProcessingStatus.Cancelled
                processing_outputs = None
                processing_err = 'Cancelled'
            elif self.tosuspend:
                self.suspended_processings.append(proc.internal_id)
                processing_status = ProcessingStatus.Suspended
                processing_outputs = None
                processing_err = 'Suspend'
            elif self.toresume:
                # self.old_processings.append(processing['processing_metadata']['internal_id'])
                # self.active_processings.clear()
                # self.active_processings.remove(processing['processing_metadata']['internal_id'])
                processing['processing_metadata']['resuming_at'] = datetime.datetime.utcnow()
                processing_status = ProcessingStatus.Running
                reset_expired_at = True
                processing_outputs = None
                processing_err = None
            else:
                processing_status = job_status
                processing_err = job_err_msg
            return processing_status, processing_outputs, processing_err, reset_expired_at
        except Exception as ex:
            self.logger.error("processing_id %s exception: %s, %s" % (processing['processing_id'], str(ex), traceback.format_exc()))
            proc.retries += 1
            if proc.retries > 10:
                processing_status = ProcessingStatus.Failed
            else:
                processing_status = ProcessingStatus.Running
            return processing_status, None, None, False

    def poll_processing_updates(self, processing, input_output_maps):
        processing_status, processing_outputs, processing_err, reset_expired_at = self.poll_processing(processing)

        # processing_metadata = processing['processing_metadata']
        # if not processing_metadata:
        #     processing_metadata = {}
        # processing_metadata['errors'] = processing_err
        proc = processing['processing_metadata']['processing']
        proc.errors = processing_err

        update_processing = {'processing_id': processing['processing_id'],
                             'parameters': {'status': processing_status,
                                            'processing_metadata': processing['processing_metadata'],
                                            'output_metadata': processing_outputs}}

        if reset_expired_at:
            update_processing['parameters']['expired_at'] = None
            processing['expired_at'] = None
        updated_contents = []
        return update_processing, updated_contents, {}



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019


"""
Class of collection lister plubin
"""

import traceback

from idds.common import exceptions
from idds.atlas.rucio.base_plugin import RucioPluginBase


class ContentsLister(RucioPluginBase):
    def __init__(self, **kwargs):
        super(ContentsLister, self).__init__(**kwargs)

    def __call__(self, scope, name):
        try:
            ret_files = []
            files = self.client.list_files(scope=scope, name=name)
            for file in files:
                ret_file = {'scope': file['scope'],
                            'name': file['name'],
                            'bytes': file['bytes'],
                            'events': file['events'],
                            'adler32': file['adler32']}
                ret_files.append(ret_file)
            return ret_files
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            raise exceptions.AgentPluginError('%s: %s' % (str(ex), traceback.format_exc()))



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020 - 2021


import json
import os
import re
import traceback

# from rucio.client.client import Client as RucioClient
from rucio.common.exception import (CannotAuthenticate as RucioCannotAuthenticate)

from idds.common import exceptions
from idds.common.constants import (ContentStatus,
                                   ProcessingStatus, WorkStatus)
from idds.common.utils import extract_scope_atlas
# from idds.workflowv2.work import Work, Processing
# from idds.workflowv2.workflow import Condition
from .atlaspandawork import ATLASPandaWork


class ATLASLocalPandaWork(ATLASPandaWork):
    def __init__(self, task_parameters=None,
                 work_tag='atlas', exec_type='panda', work_id=None,
                 primary_input_collection=None, other_input_collections=None,
                 input_collections=None,
                 primary_output_collection=None, other_output_collections=None,
                 output_collections=None, log_collections=None,
                 logger=None,
                 num_retries=5,
                 ):

        self.work_dir = "/tmp"
        self.output_files = []

        super(ATLASLocalPandaWork, self).__init__(task_parameters=task_parameters,
                                                  work_tag=work_tag,
                                                  exec_type=exec_type,
                                                  work_id=work_id,
                                                  primary_input_collection=primary_input_collection,
                                                  other_input_collections=other_input_collections,
                                                  input_collections=input_collections,
                                                  primary_output_collection=primary_output_collection,
                                                  other_output_collections=other_output_collections,
                                                  output_collections=output_collections,
                                                  log_collections=log_collections,
                                                  logger=logger,
                                                  num_retries=num_retries)

    def set_agent_attributes(self, attrs, req_attributes=None):
        if self.class_name not in attrs or 'life_time' not in attrs[self.class_name] or int(attrs[self.class_name]['life_time']) <= 0:
            attrs['life_time'] = None
        super(ATLASLocalPandaWork, self).set_agent_attributes(attrs)
        if self.agent_attributes and 'num_retries' in self.agent_attributes and self.agent_attributes['num_retries']:
            self.num_retries = int(self.agent_attributes['num_retries'])
        if self.agent_attributes and 'work_dir' in self.agent_attributes and self.agent_attributes['work_dir']:
            self.work_dir = self.agent_attributes['work_dir']

    def parse_task_parameters(self, task_parameters):
        super(ATLASLocalPandaWork, self).parse_task_parameters(task_parameters)

        try:
            if self.task_parameters and 'jobParameters' in self.task_parameters:
                jobParameters = self.task_parameters['jobParameters']
                for jobP in jobParameters:
                    if type(jobP) in [dict]:
                        if 'dataset' in jobP and 'param_type' in jobP:
                            if jobP['param_type'] == 'output' and 'value' in jobP:
                                output_c = jobP['dataset']
                                scope, name = extract_scope_atlas(output_c, scopes=[])
                                # output_coll = {'scope': scope, 'name': name}
                                output_f = jobP['value']
                                output_file = {'scope': scope, 'name': name, 'file': output_f}
                                self.output_files.append(output_file)
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))
            self.add_errors(str(ex))

    def renew_parameters_from_attributes(self):
        super(ATLASLocalPandaWork, self).renew_parameters_from_attributes()
        if not self.task_parameters:
            return

        try:
            if 'taskName' in self.task_parameters:
                self.task_name = self.task_parameters['taskName']
                self.task_name = self.renew_parameter(self.task_name)
                self.set_work_name(self.task_name)

            if 'prodSourceLabel' in self.task_parameters:
                self.task_type = self.task_parameters['prodSourceLabel']

            if 'jobParameters' in self.task_parameters:
                jobParameters = self.task_parameters['jobParameters']
                for jobP in jobParameters:
                    if type(jobP) in [dict]:
                        for key in jobP:
                            if jobP[key] and type(jobP[key]) in [str]:
                                jobP[key] = self.renew_parameter(jobP[key])
            for coll_id in self.collections:
                coll_name = self.collections[coll_id].name
                self.collections[coll_id].name = self.renew_parameter(coll_name)

            output_files = self.output_files
            self.output_files = []
            for output_file in output_files:
                output_file['name'] = self.renew_parameter(output_file['name'])
                output_file['file'] = self.renew_parameter(output_file['file'])
                self.output_files.append(output_file)
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))
            self.add_errors(str(ex))

    def set_output_data(self, data):
        self.output_data = data
        if data and type(data) in [dict]:
            for key in data:
                new_key = "user_" + str(key)
                setattr(self, new_key, data[key])

    def match_pattern_file(self, pattern, lfn):
        pattern1 = "\\$[_a-zA-Z0-9]+"
        pattern2 = "\\$\\{[_a-zA-Z0-9\\/]+\\}"
        while True:
            m = re.search(pattern1, pattern)
            if m:
                pattern = pattern.replace(m.group(0), "*")
            else:
                break
        while True:
            m = re.search(pattern2, pattern)
            if m:
                pattern = pattern.replace(m.group(0), "*")
            else:
                break

        pattern = pattern.replace(".", "\\.")
        pattern = pattern.replace("*", ".*")
        pattern = pattern + "$"

        m = re.search(pattern, lfn)
        if m:
            return True
        return False

    def ping_output_files(self):
        try:
            rucio_client = self.get_rucio_client()
            for output_i in range(len(self.output_files)):
                output_file = self.output_files[output_i]
                files = rucio_client.list_files(scope=output_file['scope'],
                                                name=output_file['name'])
                files = [f for f in files]
                self.logger.debug("ping_output_files found files for dataset(%s:%s): %s" % (output_file['scope'],
                                                                                            output_file['name'],
                                                                                            str([f['name'] for f in files])))
                for f in files:
                    if self.match_pattern_file(output_file['file'], f['name']):
                        self.output_files[output_i]['lfn'] = f['name']
            return True
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))
        return False

    def download_output_files_rucio(self, file_items):
        try:
            ret = {}
            self.logger.debug("download_output_files_rucio: %s" % str(file_items))
            if file_items:
                client = self.get_rucio_download_client()
                outputs = client.download_dids(file_items)
                for output in outputs:
                    if 'dest_file_paths' in output and output['dest_file_paths']:
                        ret[output['name']] = output['dest_file_paths'][0]
            return ret
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))
        return {}

    def get_download_dir(self, processing):
        req_dir = 'request_%s_%s/transform_%s' % (processing['request_id'],
                                                  processing['workload_id'],
                                                  processing['transform_id'])
        d_dir = os.path.join(self.work_dir, req_dir)
        if not os.path.exists(d_dir):
            os.makedirs(d_dir)
        return d_dir

    def download_output_files(self, processing):
        try:
            failed_items = []
            file_items = []
            for output_i in range(len(self.output_files)):
                if 'lfn' in self.output_files[output_i]:
                    self.logger.debug("download_output_files, Processing (%s) lfn for %s: %s" % (processing['processing_id'],
                                                                                                 self.output_files[output_i]['file'],
                                                                                                 self.output_files[output_i]['lfn']))
                    file_item = {'did': "%s:%s" % (self.output_files[output_i]['scope'], self.output_files[output_i]['lfn']),
                                 'base_dir': self.get_download_dir(processing),
                                 'no_subdir': True}
                    file_items.append(file_item)
                else:
                    self.logger.warn("download_output_files, Processing (%s) lfn for %s not found" % (processing['processing_id'],
                                                                                                      self.output_files[output_i]['file']))

            pfn_items = self.download_output_files_rucio(file_items)
            for output_i in range(len(self.output_files)):
                if 'lfn' in self.output_files[output_i]:
                    if self.output_files[output_i]['lfn'] in pfn_items:
                        pfn = pfn_items[self.output_files[output_i]['lfn']]
                        self.output_files[output_i]['pfn'] = pfn
                        self.logger.info("download_output_files, Processing (%s) pfn for %s: %s" % (processing['processing_id'],
                                                                                                    self.output_files[output_i]['file'],
                                                                                                    self.output_files[output_i]['pfn']))
                    else:
                        self.logger.info("download_output_files, Processing (%s) pfn cannot be found for %s" % (processing['processing_id'],
                                                                                                                self.output_files[output_i]['file']))
                        failed_items.append(self.output_files[output_i])
            return failed_items
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))
        return []

    def parse_output_file(self, pfn):
        try:
            if not os.path.exists(pfn):
                self.logger.warn("%s doesn't exist" % pfn)
                return {}
            else:
                with open(pfn, 'r') as f:
                    data = f.read()
                outputs = json.loads(data)
                return outputs
        except Exception as ex:
            # self.logger.error(ex)
            # self.logger.error(traceback.format_exc())
            raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))
        return {}

    def parse_output_files(self, processing):
        try:
            output_data = {}
            for output_i in range(len(self.output_files)):
                if 'pfn' in self.output_files[output_i]:
                    data = self.parse_output_file(self.output_files[output_i]['pfn'])
                    if type(data) in [dict]:
                        output_data.update(data)
            return True, output_data
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))
        return False, output_data

    def process_outputs(self, processing):
        ping_status = self.ping_output_files()
        self.logger.debug("ping_output_files(Processing_id: %s), status: %s" % (processing['processing_id'], ping_status))

        failed_items = self.download_output_files(processing)
        self.logger.debug("download_output_files(Processing_id: %s), failed_items: %s" % (processing['processing_id'], str(failed_items)))

        parse_status, output_data = self.parse_output_files(processing)
        self.logger.debug("parse_output_files(Processing_id: %s), parse_status: %s, output_data: %s" % (processing['processing_id'], parse_status, str(output_data)))

        if ping_status and not failed_items and parse_status:
            return True, output_data
        return False, output_data

    def get_rucio_download_client(self):
        try:
            from rucio.client.downloadclient import DownloadClient
            client = DownloadClient()
        except RucioCannotAuthenticate as error:
            self.logger.error(error)
            self.logger.error(traceback.format_exc())
            raise exceptions.IDDSException('%s: %s' % (str(error), traceback.format_exc()))
        return client

    def poll_panda_task_output(self, processing=None, input_output_maps=None):
        task_id = None
        try:
            from pandatools import Client

            if processing:
                output_metadata = {}
                proc = processing['processing_metadata']['processing']
                task_id = proc.workload_id
                if task_id is None:
                    task_id = self.get_panda_task_id(processing)

                if task_id:
                    # ret_ids = Client.getPandaIDsWithTaskID(task_id, verbose=False)
                    task_info = Client.getJediTaskDetails({'jediTaskID': task_id}, True, True, verbose=False)
                    self.logger.info("poll_panda_task, task_info: %s" % str(task_info))
                    if task_info[0] != 0:
                        self.logger.warn("poll_panda_task %s, error getting task status, task_info: %s" % (task_id, str(task_info)))
                        return ProcessingStatus.Submitting, [], {}

                    task_info = task_info[1]

                    processing_status = self.get_processing_status_from_panda_status(task_info["status"])

                    if processing_status in [ProcessingStatus.SubFinished]:
                        if self.retry_number < self.num_retries:
                            self.reactivate_processing(processing)
                            processing_status = ProcessingStatus.Submitted
                            self.retry_number += 1
                    if processing_status in [ProcessingStatus.SubFinished, ProcessingStatus.Finished]:
                        output_status, output_metadata = self.process_outputs(processing)
                        if not output_status:
                            err = "Failed to process processing(processing_id: %s, task_id: %s) outputs" % (processing['processing_id'], task_id)
                            self.logger.error(err)
                            self.add_errors(err)
                            processing_status = ProcessingStatus.Failed

                    return processing_status, [], {}, output_metadata
                else:
                    return ProcessingStatus.Failed, [], {}, output_metadata
        except Exception as ex:
            msg = "Failed to check the processing (%s) status: %s" % (str(processing['processing_id']), str(ex))
            self.logger.error(msg)
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.IDDSException(msg)
        return ProcessingStatus.Submitting, [], {}, {}

    def poll_processing_updates(self, processing, input_output_maps):
        """
        *** Function called by Carrier agent.
        """
        updated_contents = []
        update_processing = {}
        reset_expired_at = False
        reactive_contents = []
        # self.logger.debug("poll_processing_updates, input_output_maps: %s" % str(input_output_maps))

        if processing:
            proc = processing['processing_metadata']['processing']
            if proc.tocancel:
                self.logger.info("Cancelling processing (processing id: %s, jediTaskId: %s)" % (processing['processing_id'], proc.workload_id))
                self.kill_processing_force(processing)
                # self.kill_processing(processing)
                proc.tocancel = False
                proc.polling_retries = 0
            elif proc.tosuspend:
                self.logger.info("Suspending processing (processing id: %s, jediTaskId: %s)" % (processing['processing_id'], proc.workload_id))
                self.kill_processing_force(processing)
                # self.kill_processing(processing)
                proc.tosuspend = False
                proc.polling_retries = 0
            elif proc.toresume:
                self.logger.info("Resuming processing (processing id: %s, jediTaskId: %s)" % (processing['processing_id'], proc.workload_id))
                self.reactivate_processing(processing)
                reset_expired_at = True
                proc.toresume = False
                proc.polling_retries = 0
                proc.has_new_updates()
                # reactive_contents = self.reactive_contents(input_output_maps)
            # elif self.is_processing_expired(processing):
            elif proc.toexpire:
                self.logger.info("Expiring processing (processing id: %s, jediTaskId: %s)" % (processing['processing_id'], proc.workload_id))
                self.kill_processing(processing)
                proc.toexpire = False
                proc.polling_retries = 0
            elif proc.tofinish or proc.toforcefinish:
                self.logger.info("Finishing processing (processing id: %s, jediTaskId: %s)" % (processing['processing_id'], proc.workload_id))
                self.kill_processing(processing)
                proc.tofinish = False
                proc.toforcefinish = False
                proc.polling_retries = 0

            processing_status, poll_updated_contents, new_input_output_maps, output_metadata = self.poll_panda_task_output(processing=processing, input_output_maps=input_output_maps)
            self.logger.debug("poll_processing_updates, processing_status: %s" % str(processing_status))
            self.logger.debug("poll_processing_updates, update_contents: %s" % str(poll_updated_contents))
            self.logger.debug("poll_processing_updates, output_metadata: %s" % str(output_metadata))

            if poll_updated_contents:
                proc.has_new_updates()
            for content in poll_updated_contents:
                updated_content = {'content_id': content['content_id'],
                                   'substatus': content['substatus'],
                                   'content_metadata': content['content_metadata']}
                updated_contents.append(updated_content)

            content_substatus = {'finished': 0, 'unfinished': 0}
            for map_id in input_output_maps:
                outputs = input_output_maps[map_id]['outputs']
                for content in outputs:
                    if content.get('substatus', ContentStatus.New) != ContentStatus.Available:
                        content_substatus['unfinished'] += 1
                    else:
                        content_substatus['finished'] += 1

            if processing_status in [ProcessingStatus.SubFinished, ProcessingStatus.Finished, ProcessingStatus.Failed] and updated_contents:
                self.logger.info("Processing %s is terminated, but there are still contents to be flushed. Waiting." % (proc.workload_id))
                # there are still polling contents, should not terminate the task.
                processing_status = ProcessingStatus.Running

            if processing_status in [ProcessingStatus.SubFinished] and content_substatus['finished'] > 0 and content_substatus['unfinished'] == 0:
                # found that a 'done' panda task has got a 'finished' status. Maybe in this case 'finished' is a transparent status.
                if proc.polling_retries is None:
                    proc.polling_retries = 0

            if processing_status in [ProcessingStatus.SubFinished, ProcessingStatus.Finished, ProcessingStatus.Failed]:
                if proc.polling_retries is not None and proc.polling_retries < 3:
                    self.logger.info("processing %s polling_retries(%s) < 3, keep running" % (processing['processing_id'], proc.polling_retries))
                    processing_status = ProcessingStatus.Running
                    proc.polling_retries += 1
            else:
                proc.polling_retries = 0

            if proc.in_operation_time():
                processing_status = ProcessingStatus.Running

            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': processing_status,
                                                'output_metadata': output_metadata}}

            if reset_expired_at:
                processing['expired_at'] = None
                update_processing['parameters']['expired_at'] = None
                proc.polling_retries = 0
                # if (processing_status in [ProcessingStatus.SubFinished, ProcessingStatus.Finished, ProcessingStatus.Failed]
                #     or processing['status'] in [ProcessingStatus.Resuming]):   # noqa W503
                # using polling_retries to poll it again when panda may update the status in a delay(when issuing retryTask, panda will not update it without any delay).
                update_processing['parameters']['status'] = ProcessingStatus.Resuming
            proc.status = update_processing['parameters']['status']

        self.logger.debug("poll_processing_updates, task: %s, update_processing: %s" %
                          (proc.workload_id, str(update_processing)))
        self.logger.debug("poll_processing_updates, task: %s, updated_contents: %s" %
                          (proc.workload_id, str(updated_contents)))
        self.logger.debug("poll_processing_updates, task: %s, reactive_contents: %s" %
                          (proc.workload_id, str(reactive_contents)))
        return update_processing, updated_contents + reactive_contents, new_input_output_maps

    def syn_work_status(self, registered_input_output_maps, all_updates_flushed=True, output_statistics={}, to_release_input_contents=[]):
        super(ATLASLocalPandaWork, self).syn_work_status(registered_input_output_maps, all_updates_flushed, output_statistics, to_release_input_contents)
        # self.get_status_statistics(registered_input_output_maps)
        self.status_statistics = output_statistics

        self.logger.debug("syn_work_status, self.active_processings: %s" % str(self.active_processings))
        self.logger.debug("syn_work_status, self.has_new_inputs(): %s" % str(self.has_new_inputs))
        self.logger.debug("syn_work_status, coll_metadata_is_open: %s" %
                          str(self.collections[self._primary_input_collection].coll_metadata['is_open']))
        self.logger.debug("syn_work_status, primary_input_collection_status: %s" %
                          str(self.collections[self._primary_input_collection].status))

        self.logger.debug("syn_work_status(%s): is_processings_terminated: %s" % (str(self.get_processing_ids()), str(self.is_processings_terminated())))
        self.logger.debug("syn_work_status(%s): is_input_collections_closed: %s" % (str(self.get_processing_ids()), str(self.is_input_collections_closed())))
        self.logger.debug("syn_work_status(%s): has_new_inputs: %s" % (str(self.get_processing_ids()), str(self.has_new_inputs)))
        self.logger.debug("syn_work_status(%s): has_to_release_inputs: %s" % (str(self.get_processing_ids()), str(self.has_to_release_inputs())))
        self.logger.debug("syn_work_status(%s): to_release_input_contents: %s" % (str(self.get_processing_ids()), str(to_release_input_contents)))

        if self.is_processings_terminated() and self.is_input_collections_closed() and not self.has_new_inputs and not self.has_to_release_inputs() and not to_release_input_contents:
            # if not self.is_all_outputs_flushed(registered_input_output_maps):
            if not all_updates_flushed:
                self.logger.warn("The work processings %s is terminated. but not all outputs are flushed. Wait to flush the outputs then finish the transform" % str(self.get_processing_ids()))
                return

            if self.is_processings_finished():
                self.status = WorkStatus.Finished
            elif self.is_processings_subfinished():
                self.status = WorkStatus.SubFinished
            elif self.is_processings_failed():
                self.status = WorkStatus.Failed
            elif self.is_processings_expired():
                self.status = WorkStatus.Expired
            elif self.is_processings_cancelled():
                self.status = WorkStatus.Cancelled
            elif self.is_processings_suspended():
                self.status = WorkStatus.Suspended
        elif self.is_processings_running():
            self.status = WorkStatus.Running
        else:
            self.status = WorkStatus.Transforming

        if self.is_processings_terminated() or self.is_processings_running() or self.is_processings_started():
            self.started = True



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

import copy
import json
import os
import traceback
import uuid
import urllib
import ssl

from idds.common import exceptions
from idds.common.constants import (TransformType, CollectionType, CollectionStatus,
                                   ContentStatus, ContentType,
                                   ProcessingStatus, WorkStatus)
from idds.workflowv2.work import Work
from idds.workflowv2.workflow import Condition
import logging


class DomaCondition(Condition):
    def __init__(self, cond=None, current_work=None, true_work=None, false_work=None):
        super(DomaCondition, self).__init__(cond=cond, current_work=current_work,
                                            true_work=true_work, false_work=false_work)


class DomaLSSTWork(Work):
    def __init__(self, executable=None, arguments=None, parameters=None, setup=None,
                 work_tag='lsst', exec_type='panda', sandbox=None, work_id=None,
                 primary_input_collection=None, other_input_collections=None, input_collections=None,
                 primary_output_collection=None, other_output_collections=None,
                 output_collections=None, log_collections=None,
                 logger=None, dependency_map=None, task_name=""):
        """
        Init a work/task/transformation.

        :param setup: A string to setup the executable enviroment, it can be None.
        :param executable: The executable.
        :param arguments: The arguments.
        :param parameters: A dict with arguments needed to be replaced.
        :param work_type: The work type like data carousel, hyperparameteroptimization and so on.
        :param exec_type: The exec type like 'local', 'remote'(with remote_package set), 'docker' and so on.
        :param sandbox: The sandbox.
        :param work_id: The work/task id.
        :param primary_input_collection: The primary input collection.
        :param other_input_collections: List of the input collections.
        :param output_collections: List of the output collections.
        # :param workflow: The workflow the current work belongs to.
        """

        super(DomaLSSTWork, self).__init__(executable=executable, arguments=arguments,
                                           parameters=parameters, setup=setup, work_type=TransformType.Processing,
                                           work_tag=work_tag, exec_type=exec_type, sandbox=sandbox, work_id=work_id,
                                           primary_input_collection=primary_input_collection,
                                           other_input_collections=other_input_collections,
                                           primary_output_collection=primary_output_collection,
                                           other_output_collections=other_output_collections,
                                           input_collections=input_collections,
                                           output_collections=output_collections,
                                           log_collections=log_collections,
                                           logger=logger)
        self.pandamonitor = None
        self.dependency_map = dependency_map
        self.logger.setLevel(logging.DEBUG)
        self.task_name = task_name

    def my_condition(self):
        if self.is_finished():
            return True
        return False

    def jobs_to_idd_ds_status(self, jobstatus):
        if jobstatus == 'finished':
            return ContentStatus.Available
        elif jobstatus == 'failed ':
            return ContentStatus.Failed
        else:
            return ContentStatus.Processing

    def load_panda_config(self):
        panda_config = ConfigParser.SafeConfigParser()
        if os.environ.get('IDDS_PANDA_CONFIG', None):
            configfile = os.environ['IDDS_PANDA_CONFIG']
            if panda_config.read(configfile) == [configfile]:
                return panda_config

        configfiles = ['%s/etc/panda/panda.cfg' % os.environ.get('IDDS_HOME', ''),
                       '/etc/panda/panda.cfg', '/opt/idds/etc/panda/panda.cfg',
                       '%s/etc/panda/panda.cfg' % os.environ.get('VIRTUAL_ENV', '')]
        for configfile in configfiles:
            if panda_config.read(configfile) == [configfile]:
                return panda_config
        return panda_config

    def load_panda_monitor(self):
        panda_config = self.load_panda_config()
        self.logger.info("panda config: %s" % panda_config)
        if panda_config.has_section('panda'):
            if panda_config.has_option('panda', 'pandamonitor'):
                pandamonitor = panda_config.get('panda', 'pandamonitor')
                return pandamonitor
        return None

    def poll_external_collection(self, coll):
        try:
            # if 'coll_metadata' in coll and 'is_open' in coll['coll_metadata'] and not coll['coll_metadata']['is_open']:
            if 'status' in coll and coll['status'] in [CollectionStatus.Closed]:
                return coll
            else:
                # client = self.get_rucio_client()
                # did_meta = client.get_metadata(scope=coll['scope'], name=coll['name'])
                if 'coll_metadata' not in coll:
                    coll['coll_metadata'] = {}
                coll['coll_metadata']['bytes'] = 1
                # coll['coll_metadata']['total_files'] = 1
                coll['coll_metadata']['availability'] = 1
                coll['coll_metadata']['events'] = 1
                coll['coll_metadata']['is_open'] = True
                coll['coll_metadata']['run_number'] = 1
                coll['coll_metadata']['did_type'] = 'DATASET'
                coll['coll_metadata']['list_all_files'] = False

                if 'is_open' in coll['coll_metadata'] and not coll['coll_metadata']['is_open']:
                    coll_status = CollectionStatus.Closed
                else:
                    coll_status = CollectionStatus.Open
                coll['status'] = coll_status
                coll['coll_type'] = CollectionType.Dataset

                return coll
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))

    def get_input_collections(self):
        """
        *** Function called by Transformer agent.
        """
        colls = [self._primary_input_collection] + self._other_input_collections
        for coll_int_id in colls:
            coll = self.collections[coll_int_id]
            coll = self.poll_external_collection(coll)
            self.collections[coll_int_id] = coll
        return super(DomaLSSTWork, self).get_input_collections()

    def get_unsubmitted_inputs(self):
        not_submitted_inputs = filter(lambda t: not t["submitted"], self.dependency_map)
        tasks_to_check = []
        for job in not_submitted_inputs:
            tasks_to_check.extend([(input["task"], input["inputname"]) for input in job["dependencies"] if not input["available"]])
        tasks_to_check_compact = {}
        for task in tasks_to_check:
            tasks_to_check_compact.setdefault(task[0], set()).add(task[1])
        return tasks_to_check_compact

    def set_dependency_input_available(self, taskname, inputname):
        for job in self.dependency_map:
            for dependency in job["dependencies"]:
                if dependency["task"] == taskname and dependency["inputname"] == inputname:
                    dependency["available"] = True

    def update_dependencies(self):
        tasks_to_check = self.get_unsubmitted_inputs()
        for task, inputs in tasks_to_check.items():
            _, outputs = self.poll_panda_task(task_name=task)
            for input in inputs:
                if outputs.get(input, ContentStatus.Processing) == ContentStatus.Available:
                    self.set_dependency_input_available(task, input)

    def get_ready_inputs(self):
        not_submitted_inputs = filter(lambda j: not j["submitted"], self.dependency_map)
        files_to_submit = []
        for job in not_submitted_inputs:
            unresolved_deps = [input for input in job["dependencies"] if not input["available"]]
            if len(unresolved_deps) == 0:
                files_to_submit.append(job["name"])
        return files_to_submit

    def check_dependencies(self):
        self.update_dependencies()
        return self.get_ready_inputs()

    def can_close(self):
        not_submitted_inputs = list(filter(lambda t: not t["submitted"], self.dependency_map))
        if len(not_submitted_inputs) == 0:
            return True
        else:
            return False

    def get_input_contents(self):
        """
        Get all input contents from DDM.
        """
        try:
            files = self.check_dependencies()
            ret_files = []
            coll = self.collections[self._primary_input_collection]
            for file in files:
                ret_file = {'coll_id': coll['coll_id'],
                            'scope': coll['scope'],
                            'name': file,  # or a different file name from the dataset name
                            'bytes': 1,
                            'adler32': '12345678',
                            'min_id': 0,
                            'max_id': 1,
                            'content_type': ContentType.File,
                            'content_metadata': {'events': 1}}  # here events is all events for eventservice, not used here.
                ret_files.append(ret_file)
            return ret_files
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))

    def get_mapped_inputs(self, mapped_input_output_maps):
        ret = []
        for map_id in mapped_input_output_maps:
            inputs = mapped_input_output_maps[map_id]['inputs']

            # if 'primary' is not set, the first one is the primary input.
            primary_input = inputs[0]
            for ip in inputs:
                if 'primary' in ip['content_metadata'] and ip['content_metadata']['primary']:
                    primary_input = ip
            ret.append(primary_input)
        return ret

    def get_new_input_output_maps(self, mapped_input_output_maps={}):
        """
        *** Function called by Transformer agent.
        New inputs which are not yet mapped to outputs.

        :param mapped_input_output_maps: Inputs that are already mapped.
        """
        inputs = self.get_input_contents()
        mapped_inputs = self.get_mapped_inputs(mapped_input_output_maps)
        mapped_inputs_scope_name = [ip['name'] for ip in mapped_inputs]

        new_inputs = []
        new_input_output_maps = {}
        for ip in inputs:
            ip_scope_name = ip['name']
            if ip_scope_name not in mapped_inputs_scope_name:
                new_inputs.append(ip)

        # to avoid cheking new inputs if there are no new inputs anymore
        if not new_inputs and self.collections[self._primary_input_collection]['status'] in [CollectionStatus.Closed]:
            self.set_has_new_inputs(False)
        else:
            mapped_keys = mapped_input_output_maps.keys()
            if mapped_keys:
                next_key = max(mapped_keys) + 1
            else:
                next_key = 1
            for ip in new_inputs:
                out_ip = copy.deepcopy(ip)
                out_ip['coll_id'] = self.collections[self._primary_output_collection]['coll_id']
                new_input_output_maps[next_key] = {'inputs': [ip],
                                                   'outputs': [out_ip]}
                next_key += 1
        self.logger.debug("get_new_input_output_maps, new_input_output_maps: %s" % str(new_input_output_maps))

        for index, _inputs in new_input_output_maps.items():
            if len(_inputs['inputs']) > 0:
                for item in self.dependency_map:
                    if item["name"] == _inputs['inputs'][0]["name"]:
                        item["submitted"] = True  # 0 is used due to a single file pseudo input

        if self.can_close():
            self.collections[self._primary_input_collection]['coll_metadata']['is_open'] = False
            self.collections[self._primary_input_collection]['status'] = CollectionStatus.Closed

        return new_input_output_maps

    def get_processing(self, input_output_maps):
        """
        *** Function called by Transformer agent.

        If there is already an active processing for this work, will do nothing.
        If there is no active processings, create_processing will be called.
        """
        if self.active_processings:
            return self.processings[self.active_processings[0]]
        else:
            return None

    def create_processing(self, input_output_maps):
        """
        *** Function called by Transformer agent.

        :param input_output_maps: new maps from inputs to outputs.
        """
        in_files = []
        for map_id in input_output_maps:
            # one map is a job which transform the inputs to outputs.
            inputs = input_output_maps[map_id]['inputs']
            # outputs = input_output_maps[map_id]['outputs']
            for ip in inputs:
                in_files.append(ip['name'])

        taskParamMap = {}
        taskParamMap['vo'] = 'wlcg'
        taskParamMap['site'] = 'BNL_OSG_1'
        taskParamMap['workingGroup'] = 'lsst'
        taskParamMap['nFilesPerJob'] = 1
        taskParamMap['nFiles'] = len(in_files)
        taskParamMap['noInput'] = True
        taskParamMap['pfnList'] = in_files
        taskParamMap['taskName'] = self.task_name
        taskParamMap['userName'] = 'Siarhei Padolski'
        taskParamMap['taskPriority'] = 900
        taskParamMap['architecture'] = ''
        taskParamMap['transUses'] = ''
        taskParamMap['transHome'] = None
        taskParamMap['transPath'] = 'https://atlpan.web.cern.ch/atlpan/bash-c'
        taskParamMap['processingType'] = 'testidds'
        taskParamMap['prodSourceLabel'] = 'test'
        taskParamMap['taskType'] = 'test'
        taskParamMap['coreCount'] = 1
        taskParamMap['skipScout'] = True
        taskParamMap['cloud'] = 'US'
        taskParamMap['jobParameters'] = [
            {'type': 'constant',
             'value': "echo ${IN/L}",  # noqa: E501
             },
        ]

        proc = {'processing_metadata': {'internal_id': str(uuid.uuid1()),
                                        'task_id': None,
                                        'task_param': taskParamMap}}
        self.add_processing_to_processings(proc)
        self.active_processings.append(proc['processing_metadata']['internal_id'])
        return proc

    def submit_panda_task(self, processing):
        try:
            from pandatools import Client

            task_param = processing['processing_metadata']['task_param']
            return_code = Client.insertTaskParams(task_param, verbose=True)
            if return_code[0] == 0:
                return return_code[1][1]
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.AgentPluginError('%s: %s' % (str(ex), traceback.format_exc()))
        return None

    def submit_processing(self, processing):
        """
        *** Function called by Carrier agent.
        """
        if 'task_id' in processing['processing_metadata'] and processing['processing_metadata']['task_id']:
            pass
        else:
            task_id = self.submit_panda_task(processing)
            processing['processing_metadata']['task_id'] = task_id

    def download_payload_json(self, task_url):
        response = None
        try:
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
            req = urllib.request.Request(task_url)
            response = urllib.request.urlopen(req, timeout=180, context=ctx).read()
            response = json.loads(response)
        except Exception or urllib.request.error as e:
            raise e
        return response

    def poll_panda_task(self, processing=None, task_name=None):
        try:
            if not self.pandamonitor:
                self.pandamonitor = self.load_panda_monitor()
                self.logger.info("panda server: %s" % self.pandamonitor)

            task_id = None
            task_info = None
            if processing:
                task_id = processing['processing_metadata']['task_id']
                task_url = self.pandamonitor + '/task/?json&jeditaskid=' + str(task_id)
                task_info = self.download_payload_json(task_url)
            elif task_name:
                task_url = self.pandamonitor + '/tasks/?taskname=' + str(task_name) + "&json"
                self.logger.debug("poll_panda_task, task_url: %s" % str(task_url))
                self.logger.debug("poll_panda_task, task_url: %s" % str(self.download_payload_json(task_url)))
                task_json = self.download_payload_json(task_url)
                if len(task_json) > 0:
                    task_info = task_json[0]
                else:
                    return "No status", {}
            if not task_id:
                task_id = task_info.get('jeditaskid', None)
                if not task_id:
                    return "No status", {}

            jobs_url = self.pandamonitor + '/jobs/?json&datasets=yes&jeditaskid=' + str(task_id)
            jobs_list = self.download_payload_json(jobs_url)
            outputs_status = {}
            for job_info in jobs_list['jobs']:
                if 'jobstatus' in job_info and 'datasets' in job_info and len(job_info['datasets']) > 0:
                    output_index = job_info['datasets'][0]['lfn'].split(':')[1]
                    status = self.jobs_to_idd_ds_status(job_info['jobstatus'])
                    outputs_status[output_index] = status

            task_status = None
            self.logger.debug("poll_panda_task, task_info: %s" % str(task_info))
            if task_info.get("task", None) is not None:
                task_status = task_info["task"]["status"]
            return task_status, outputs_status
        except Exception as ex:
            msg = "Failed to check the panda task(%s) status: %s" % (str(task_id), str(ex))
            raise exceptions.IDDSException(msg)

    def poll_processing_updates(self, processing, input_output_maps):
        """
        *** Function called by Carrier agent.
        """
        updated_contents = []
        update_processing = {}
        self.logger.debug("poll_processing_updates, input_output_maps: %s" % str(input_output_maps))

        if processing:
            task_status, outputs_status = self.poll_panda_task(processing=processing)

            self.logger.debug("poll_processing_updates, outputs_status: %s" % str(outputs_status))
            self.logger.debug("poll_processing_updates, task_status: %s" % str(task_status))

            content_substatus = {'finished': 0, 'unfinished': 0}
            for map_id in input_output_maps:
                outputs = input_output_maps[map_id]['outputs']
                for content in outputs:
                    key = content['name']
                    if key in outputs_status:
                        if content.get('substatus', ContentStatus.New) != outputs_status[key]:
                            updated_content = {'content_id': content['content_id'],
                                               'substatus': outputs_status[key]}
                            updated_contents.append(updated_content)
                            content['substatus'] = outputs_status[key]
                    if content['substatus'] == ContentStatus.Available:
                        content_substatus['finished'] += 1
                    else:
                        content_substatus['unfinished'] += 1

            if task_status and task_status == 'done' and content_substatus['finished'] > 0 and content_substatus['unfinished'] == 0:
                update_processing = {'processing_id': processing['processing_id'],
                                     'parameters': {'status': ProcessingStatus.Finished}}

        self.logger.debug("poll_processing_updates, update_processing: %s" % str(update_processing))
        self.logger.debug("poll_processing_updates, updated_contents: %s" % str(updated_contents))
        return update_processing, updated_contents, {}

    def get_status_statistics(self, registered_input_output_maps):
        status_statistics = {}
        for map_id in registered_input_output_maps:
            outputs = registered_input_output_maps[map_id]['outputs']

            for content in outputs:
                if content['status'].name not in status_statistics:
                    status_statistics[content['status'].name] = 0
                status_statistics[content['status'].name] += 1
        self.status_statistics = status_statistics
        self.logger.debug("registered_input_output_maps, status_statistics: %s" % str(status_statistics))
        return status_statistics

    def syn_work_status(self, registered_input_output_maps):
        self.get_status_statistics(registered_input_output_maps)
        self.logger.debug("syn_work_status, self.active_processings: %s" % str(self.active_processings))
        self.logger.debug("syn_work_status, self.has_new_inputs(): %s" % str(self.has_new_inputs()))
        self.logger.debug("syn_work_status, coll_metadata_is_open: %s" % str(self.collections[self.primary_input_collection]['coll_metadata']['is_open']))
        self.logger.debug("syn_work_status, primary_input_collection_status: %s" % str(self.collections[self.primary_input_collection]['status']))

        if self.is_processings_terminated() and not self.has_new_inputs():
            keys = self.status_statistics.keys()
            if ContentStatus.New.name in keys or ContentStatus.Processing.name in keys:
                pass
            else:
                if len(keys) == 1:
                    if ContentStatus.Available.name in keys:
                        self.status = WorkStatus.Finished
                    else:
                        self.status = WorkStatus.Failed
                else:
                    self.status = WorkStatus.SubFinished



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020 - 2021


try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

import datetime
import os
import random
import traceback

from rucio.client.client import Client as RucioClient
from rucio.common.exception import (CannotAuthenticate as RucioCannotAuthenticate)

from idds.common import exceptions
from idds.common.constants import (TransformType, CollectionStatus, CollectionType,
                                   ContentStatus, ContentType,
                                   ProcessingStatus, WorkStatus)
from idds.common.utils import extract_scope_atlas
from idds.workflowv2.work import Work, Processing
# from idds.workflowv2.workflow import Condition


class ATLASPandaWork(Work):
    def __init__(self, task_parameters=None,
                 work_tag='atlas', exec_type='panda', work_id=None,
                 primary_input_collection=None, other_input_collections=None,
                 input_collections=None,
                 primary_output_collection=None, other_output_collections=None,
                 output_collections=None, log_collections=None,
                 logger=None,
                 # dependency_map=None, task_name="",
                 # task_queue=None, processing_type=None,
                 # prodSourceLabel='test', task_type='test',
                 # maxwalltime=90000, maxattempt=5, core_count=1,
                 # encode_command_line=False,
                 num_retries=5,
                 # task_log=None,
                 # task_cloud=None,
                 # task_rss=0
                 ):

        super(ATLASPandaWork, self).__init__(work_type=TransformType.Processing,
                                             work_tag=work_tag,
                                             exec_type=exec_type,
                                             work_id=work_id,
                                             primary_input_collection=primary_input_collection,
                                             other_input_collections=other_input_collections,
                                             primary_output_collection=primary_output_collection,
                                             other_output_collections=other_output_collections,
                                             input_collections=input_collections,
                                             output_collections=output_collections,
                                             log_collections=log_collections,
                                             release_inputs_after_submitting=True,
                                             logger=logger)
        self.panda_url = None
        self.panda_url_ssl = None
        self.panda_monitor = None

        self.task_type = 'test'
        self.task_parameters = None
        self.parse_task_parameters(task_parameters)
        # self.logger.setLevel(logging.DEBUG)

        self.retry_number = 0
        self.num_retries = num_retries

        self.load_panda_urls()

    def my_condition(self):
        if self.is_finished():
            return True
        return False

    def load_panda_config(self):
        panda_config = ConfigParser.SafeConfigParser()
        if os.environ.get('IDDS_PANDA_CONFIG', None):
            configfile = os.environ['IDDS_PANDA_CONFIG']
            if panda_config.read(configfile) == [configfile]:
                return panda_config

        configfiles = ['%s/etc/panda/panda.cfg' % os.environ.get('IDDS_HOME', ''),
                       '/etc/panda/panda.cfg', '/opt/idds/etc/panda/panda.cfg',
                       '%s/etc/panda/panda.cfg' % os.environ.get('VIRTUAL_ENV', '')]
        for configfile in configfiles:
            if panda_config.read(configfile) == [configfile]:
                return panda_config
        return panda_config

    def load_panda_urls(self):
        panda_config = self.load_panda_config()
        # self.logger.debug("panda config: %s" % panda_config)
        self.panda_url = None
        self.panda_url_ssl = None
        self.panda_monitor = None

        if panda_config.has_section('panda'):
            if panda_config.has_option('panda', 'panda_monitor_url'):
                self.panda_monitor = panda_config.get('panda', 'panda_monitor_url')
                os.environ['PANDA_MONITOR_URL'] = self.panda_monitor
                # self.logger.debug("Panda monitor url: %s" % str(self.panda_monitor))
            if panda_config.has_option('panda', 'panda_url'):
                self.panda_url = panda_config.get('panda', 'panda_url')
                os.environ['PANDA_URL'] = self.panda_url
                # self.logger.debug("Panda url: %s" % str(self.panda_url))
            if panda_config.has_option('panda', 'panda_url_ssl'):
                self.panda_url_ssl = panda_config.get('panda', 'panda_url_ssl')
                os.environ['PANDA_URL_SSL'] = self.panda_url_ssl
                # self.logger.debug("Panda url ssl: %s" % str(self.panda_url_ssl))

        if not self.panda_monitor and 'PANDA_MONITOR_URL' in os.environ and os.environ['PANDA_MONITOR_URL']:
            self.panda_monitor = os.environ['PANDA_MONITOR_URL']
            # self.logger.debug("Panda monitor url: %s" % str(self.panda_monitor))
        if not self.panda_url and 'PANDA_URL' in os.environ and os.environ['PANDA_URL']:
            self.panda_url = os.environ['PANDA_URL']
            # self.logger.debug("Panda url: %s" % str(self.panda_url))
        if not self.panda_url_ssl and 'PANDA_URL_SSL' in os.environ and os.environ['PANDA_URL_SSL']:
            self.panda_url_ssl = os.environ['PANDA_URL_SSL']
            # self.logger.debug("Panda url ssl: %s" % str(self.panda_url_ssl))

    def set_agent_attributes(self, attrs, req_attributes=None):
        if self.class_name not in attrs or 'life_time' not in attrs[self.class_name] or int(attrs[self.class_name]['life_time']) <= 0:
            attrs['life_time'] = None
        super(ATLASPandaWork, self).set_agent_attributes(attrs)
        if self.agent_attributes and 'num_retries' in self.agent_attributes and self.agent_attributes['num_retries']:
            self.num_retries = int(self.agent_attributes['num_retries'])

    def parse_task_parameters(self, task_parameters):
        if self.task_parameters:
            return
        elif not task_parameters:
            return
        self.task_parameters = task_parameters

        try:
            if 'taskName' in self.task_parameters:
                self.task_name = self.task_parameters['taskName']
                self.set_work_name(self.task_name)

            if 'prodSourceLabel' in self.task_parameters:
                self.task_type = self.task_parameters['prodSourceLabel']

            if 'jobParameters' in self.task_parameters:
                jobParameters = self.task_parameters['jobParameters']
                for jobP in jobParameters:
                    if type(jobP) in [dict]:
                        if 'dataset' in jobP and 'param_type' in jobP:
                            if jobP['param_type'] == 'input':
                                input_c = jobP['dataset']
                                scope, name = extract_scope_atlas(input_c, scopes=[])
                                input_coll = {'scope': scope, 'name': name}
                                self.set_primary_input_collection(input_coll)
                            if jobP['param_type'] == 'output':
                                output_c = jobP['dataset']
                                scope, name = extract_scope_atlas(output_c, scopes=[])
                                output_coll = {'scope': scope, 'name': name}
                                self.add_output_collections([output_coll])

            if 'log' in self.task_parameters:
                log = self.task_parameters['log']
                dataset = log['dataset']
                scope, name = extract_scope_atlas(dataset, scopes=[])
                log_col = {'scope': scope, 'name': name}
                self.add_log_collections(log_col)

            if not self.get_primary_input_collection():
                output_colls = self.get_output_collections()
                output_coll = output_colls[0]
                name = 'pseudo_input.' + datetime.datetime.utcnow().strftime("%Y_%m_%d_%H_%M_%S_%f") + str(random.randint(1, 1000))
                input_coll = {'scope': output_coll.scope, 'name': name, 'type': CollectionType.PseudoDataset}
                self.set_primary_input_collection(input_coll)

        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))
            self.add_errors(str(ex))

    def renew_parameter(self, parameter):
        new_parameter = parameter
        if '___idds___' in parameter:
            pos_start = parameter.find('___idds___')
            attr = parameter[pos_start:]
            attr = attr.replace("___idds___", "")
            pos = attr.find("___")
            if pos > -1:
                attr = attr[:pos]
            if attr:
                idds_attr = "___idds___" + attr + "___"
                if hasattr(self, attr):
                    attr_value = getattr(self, attr)
                    new_parameter = parameter.replace(idds_attr, str(attr_value))
        return new_parameter

    def renew_parameters_from_attributes(self):
        if not self.task_parameters:
            return

        try:
            for key in self.task_parameters:
                if self.task_parameters[key] and type(self.task_parameters[key]) in [str]:
                    self.task_parameters[key] = self.renew_parameter(self.task_parameters[key])

            if 'taskName' in self.task_parameters:
                self.task_name = self.task_parameters['taskName']
                self.task_name = self.renew_parameter(self.task_name)
                self.task_parameters['taskName'] = self.task_name
                self.set_work_name(self.task_name)

            if 'prodSourceLabel' in self.task_parameters:
                self.task_type = self.task_parameters['prodSourceLabel']

            if 'jobParameters' in self.task_parameters:
                jobParameters = self.task_parameters['jobParameters']
                for jobP in jobParameters:
                    if type(jobP) in [dict]:
                        for key in jobP:
                            if jobP[key] and type(jobP[key]) in [str]:
                                jobP[key] = self.renew_parameter(jobP[key])

            if 'log' in self.task_parameters:
                log = self.task_parameters['log']
                for key in log:
                    if log[key] and type(log[key]) in [str]:
                        self.task_parameters['log'][key] = self.renew_parameter(log[key])

            for coll_id in self.collections:
                coll_name = self.collections[coll_id].name
                self.collections[coll_id].name = self.renew_parameter(coll_name)
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))
            self.add_errors(str(ex))

    def get_rucio_client(self):
        try:
            client = RucioClient()
        except RucioCannotAuthenticate as error:
            self.logger.error(error)
            self.logger.error(traceback.format_exc())
            raise exceptions.IDDSException('%s: %s' % (str(error), traceback.format_exc()))
        return client

    def poll_external_collection(self, coll):
        try:
            if coll.status in [CollectionStatus.Closed]:
                return coll
            else:
                try:
                    if not coll.coll_type == CollectionType.PseudoDataset:
                        client = self.get_rucio_client()
                        did_meta = client.get_metadata(scope=coll.scope, name=coll.name)

                        coll.coll_metadata['bytes'] = did_meta['bytes']
                        coll.coll_metadata['total_files'] = did_meta['length']
                        coll.coll_metadata['availability'] = did_meta['availability']
                        coll.coll_metadata['events'] = did_meta['events']
                        coll.coll_metadata['is_open'] = did_meta['is_open']
                        coll.coll_metadata['run_number'] = did_meta['run_number']
                        coll.coll_metadata['did_type'] = did_meta['did_type']
                        coll.coll_metadata['list_all_files'] = False

                        if 'is_open' in coll.coll_metadata and not coll.coll_metadata['is_open']:
                            coll_status = CollectionStatus.Closed
                        else:
                            coll_status = CollectionStatus.Open
                        coll.status = coll_status

                        if 'did_type' in coll.coll_metadata:
                            if coll.coll_metadata['did_type'] == 'DATASET':
                                coll_type = CollectionType.Dataset
                            elif coll.coll_metadata['did_type'] == 'CONTAINER':
                                coll_type = CollectionType.Container
                            else:
                                coll_type = CollectionType.File
                        else:
                            coll_type = CollectionType.Dataset
                        coll.coll_metadata['coll_type'] = coll_type
                        coll.coll_type = coll_type
                        return coll
                except Exception as ex:
                    self.logger.warn("Faield to get the dataset information(%s:%s) from Panda: %s" % (coll.scope, coll.name, str(ex)))

                coll.coll_metadata['bytes'] = 0
                coll.coll_metadata['total_files'] = 0
                coll.coll_metadata['availability'] = True
                coll.coll_metadata['events'] = 0
                coll.coll_metadata['is_open'] = False
                coll.coll_metadata['run_number'] = None
                coll.coll_metadata['did_type'] = 'DATASET'
                coll.coll_metadata['list_all_files'] = False

                if 'is_open' in coll.coll_metadata and not coll.coll_metadata['is_open']:
                    coll_status = CollectionStatus.Closed
                else:
                    coll_status = CollectionStatus.Open
                coll.status = coll_status

                coll.coll_metadata['coll_type'] = coll.coll_type

                return coll
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))

    def get_input_collections(self):
        """
        *** Function called by Transformer agent.
        """
        colls = [self._primary_input_collection] + self._other_input_collections
        for coll_int_id in colls:
            coll = self.collections[coll_int_id]
            # if self.is_internal_collection(coll):
            #     coll = self.poll_internal_collection(coll)
            # else:
            #     coll = self.poll_external_collection(coll)
            coll = self.poll_external_collection(coll)
            self.collections[coll_int_id] = coll
        return super(ATLASPandaWork, self).get_input_collections()

    def get_input_contents(self):
        """
        Get all input contents from DDM.
        """
        try:
            ret_files = []
            return ret_files
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            raise exceptions.IDDSException('%s: %s' % (str(ex), traceback.format_exc()))

    def get_mapped_inputs(self, mapped_input_output_maps):
        ret = []
        for map_id in mapped_input_output_maps:
            inputs = mapped_input_output_maps[map_id]['inputs']

            # if 'primary' is not set, the first one is the primary input.
            primary_input = inputs[0]
            for ip in inputs:
                if 'primary' in ip['content_metadata'] and ip['content_metadata']['primary']:
                    primary_input = ip
            ret.append(primary_input)
        return ret

    def get_mapped_outputs(self, mapped_input_output_maps):
        ret = []
        for map_id in mapped_input_output_maps:
            outputs = mapped_input_output_maps[map_id]['outputs']

            # if 'primary' is not set, the first one is the primary input.
            primary_output = outputs[0]
            for ip in outputs:
                if 'primary' in ip['content_metadata'] and ip['content_metadata']['primary']:
                    primary_output = ip
            ret.append(primary_output)
        return ret

    def map_file_to_content(self, coll_id, scope, name):
        content = {'coll_id': coll_id,
                   'scope': scope,
                   'name': name,  # or a different file name from the dataset name
                   'bytes': 1,
                   'adler32': '12345678',
                   'min_id': 0,
                   'max_id': 1,
                   'content_type': ContentType.File,
                   # 'content_relation_type': content_relation_type,
                   # here events is all events for eventservice, not used here.
                   'content_metadata': {'events': 1}}
        return content

    def get_next_map_id(self, input_output_maps):
        mapped_keys = input_output_maps.keys()
        if mapped_keys:
            next_key = max(mapped_keys) + 1
        else:
            next_key = 1
        return next_key

    def get_new_input_output_maps(self, mapped_input_output_maps={}):
        """
        *** Function called by Transformer agent.
        New inputs which are not yet mapped to outputs.

        :param mapped_input_output_maps: Inputs that are already mapped.
        """
        inputs = self.get_input_contents()
        mapped_inputs = self.get_mapped_inputs(mapped_input_output_maps)
        mapped_inputs_scope_name = [ip['scope'] + ":" + ip['name'] for ip in mapped_inputs]

        new_inputs = []
        new_input_output_maps = {}
        for ip in inputs:
            ip_scope_name = ip['scope'] + ":" + ip['name']
            if ip_scope_name not in mapped_inputs_scope_name:
                new_inputs.append(ip)

        # to avoid cheking new inputs if there are no new inputs anymore
        if (not new_inputs and self.collections[self._primary_input_collection].status in [CollectionStatus.Closed]):  # noqa: W503
            self.set_has_new_inputs(False)
        else:
            pass

        # self.logger.debug("get_new_input_output_maps, new_input_output_maps: %s" % str(new_input_output_maps))
        self.logger.debug("get_new_input_output_maps, new_input_output_maps len: %s" % len(new_input_output_maps))
        return new_input_output_maps

    def use_dependency_to_release_jobs(self):
        """
        *** Function called by Transformer agent.
        """
        return False

    def get_processing(self, input_output_maps=[], without_creating=False):
        """
        *** Function called by Transformer agent.

        If there is already an active processing for this work, will do nothing.
        If there is no active processings, create_processing will be called.
        """
        if self.active_processings:
            return self.processings[self.active_processings[0]]
        else:
            if not without_creating:
                # return None
                return self.create_processing(input_output_maps)
        return None

    def create_processing(self, input_output_maps=[]):
        """
        *** Function called by Transformer agent.

        :param input_output_maps: new maps from inputs to outputs.
        """
        processing_metadata = {'task_param': self.task_parameters}
        proc = Processing(processing_metadata=processing_metadata)
        proc.workload_id = None
        self.add_processing_to_processings(proc)
        self.active_processings.append(proc.internal_id)
        return proc

    def submit_panda_task(self, processing):
        try:
            from pandatools import Client

            proc = processing['processing_metadata']['processing']
            task_param = proc.processing_metadata['task_param']
            return_code = Client.insertTaskParams(task_param, verbose=True)
            if return_code[0] == 0:
                try:
                    task_id = int(return_code[1][1])
                    return task_id
                except Exception as ex:
                    self.logger.warn("task id is not retruned: (%s) is not task id: %s" % (return_code[1][1], str(ex)))
                    # jediTaskID=26468582
                    if return_code[1][1] and 'jediTaskID=' in return_code[1][1]:
                        parts = return_code[1][1].split(" ")
                        for part in parts:
                            if 'jediTaskID=' in part:
                                task_id = int(part.split("=")[1])
                                return task_id
            else:
                self.logger.warn("submit_panda_task, return_code: %s" % str(return_code))
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.AgentPluginError('%s: %s' % (str(ex), traceback.format_exc()))
        return None

    def submit_processing(self, processing):
        """
        *** Function called by Carrier agent.
        """
        proc = processing['processing_metadata']['processing']
        if proc.workload_id:
            # if 'task_id' in processing['processing_metadata'] and processing['processing_metadata']['task_id']:
            pass
        else:
            task_id = self.submit_panda_task(processing)
            # processing['processing_metadata']['task_id'] = task_id
            # processing['processing_metadata']['workload_id'] = task_id
            proc.workload_id = task_id
            if task_id:
                proc.submitted_at = datetime.datetime.utcnow()

    def poll_panda_task_status(self, processing):
        if 'processing' in processing['processing_metadata']:
            from pandatools import Client

            proc = processing['processing_metadata']['processing']
            status, task_status = Client.getTaskStatus(proc.workload_id)
            if status == 0:
                return task_status
        else:
            return 'failed'
        return None

    def get_processing_status_from_panda_status(self, task_status):
        if task_status in ['registered', 'defined', 'assigning']:
            processing_status = ProcessingStatus.Submitting
        elif task_status in ['ready', 'pending', 'scouting', 'scouted', 'prepared', 'topreprocess', 'preprocessing']:
            processing_status = ProcessingStatus.Submitted
        elif task_status in ['running', 'toretry', 'toincexec', 'throttled']:
            processing_status = ProcessingStatus.Running
        elif task_status in ['done']:
            processing_status = ProcessingStatus.Finished
        elif task_status in ['finished', 'paused']:
            # finished, finishing, waiting it to be done
            processing_status = ProcessingStatus.SubFinished
        elif task_status in ['failed', 'aborted', 'broken', 'exhausted']:
            # aborting, tobroken
            processing_status = ProcessingStatus.Failed
        else:
            # finished, finishing, aborting, topreprocess, preprocessing, tobroken
            # toretry, toincexec, rerefine, paused, throttled, passed
            processing_status = ProcessingStatus.Submitted
        return processing_status

    def get_panda_task_id(self, processing):
        from pandatools import Client

        start_time = datetime.datetime.utcnow() - datetime.timedelta(hours=10)
        start_time = start_time.strftime('%Y-%m-%d %H:%M:%S')
        status, results = Client.getJobIDsJediTasksInTimeRange(start_time, task_type=self.task_type, verbose=False)
        if status != 0:
            self.logger.warn("Error to poll latest tasks in last ten hours: %s, %s" % (status, results))
            return None

        proc = processing['processing_metadata']['processing']
        task_id = None
        for req_id in results:
            task_name = results[req_id]['taskName']
            if proc.workload_id is None and task_name == self.task_name:
                task_id = results[req_id]['jediTaskID']
                # processing['processing_metadata']['task_id'] = task_id
                # processing['processing_metadata']['workload_id'] = task_id
                proc.workload_id = task_id
                if task_id:
                    proc.submitted_at = datetime.datetime.utcnow()

        return task_id

    def poll_panda_task(self, processing=None, input_output_maps=None):
        task_id = None
        try:
            from pandatools import Client

            if processing:
                proc = processing['processing_metadata']['processing']
                task_id = proc.workload_id
                if task_id is None:
                    task_id = self.get_panda_task_id(processing)

                if task_id:
                    # ret_ids = Client.getPandaIDsWithTaskID(task_id, verbose=False)
                    task_info = Client.getJediTaskDetails({'jediTaskID': task_id}, True, True, verbose=False)
                    self.logger.info("poll_panda_task, task_info: %s" % str(task_info))
                    if task_info[0] != 0:
                        self.logger.warn("poll_panda_task %s, error getting task status, task_info: %s" % (task_id, str(task_info)))
                        return ProcessingStatus.Submitting, [], {}

                    task_info = task_info[1]

                    processing_status = self.get_processing_status_from_panda_status(task_info["status"])

                    if processing_status in [ProcessingStatus.SubFinished]:
                        if self.retry_number < self.num_retries:
                            self.reactivate_processing(processing)
                            processing_status = ProcessingStatus.Submitted
                            self.retry_number += 1

                    return processing_status, [], {}
                else:
                    return ProcessingStatus.Failed, [], {}
        except Exception as ex:
            msg = "Failed to check the processing (%s) status: %s" % (str(processing['processing_id']), str(ex))
            self.logger.error(msg)
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            # raise exceptions.IDDSException(msg)
        return ProcessingStatus.Submitting, [], {}

    def kill_processing(self, processing):
        try:
            if processing:
                from pandatools import Client
                proc = processing['processing_metadata']['processing']
                task_id = proc.workload_id
                # task_id = processing['processing_metadata']['task_id']
                # Client.killTask(task_id)
                Client.finishTask(task_id, soft=False)
        except Exception as ex:
            msg = "Failed to check the processing (%s) status: %s" % (str(processing['processing_id']), str(ex))
            raise exceptions.IDDSException(msg)

    def kill_processing_force(self, processing):
        try:
            if processing:
                from pandatools import Client
                proc = processing['processing_metadata']['processing']
                task_id = proc.workload_id
                # task_id = processing['processing_metadata']['task_id']
                Client.killTask(task_id)
                # Client.finishTask(task_id, soft=True)
        except Exception as ex:
            msg = "Failed to check the processing (%s) status: %s" % (str(processing['processing_id']), str(ex))
            raise exceptions.IDDSException(msg)

    def reactivate_processing(self, processing):
        try:
            if processing:
                from pandatools import Client
                # task_id = processing['processing_metadata']['task_id']
                proc = processing['processing_metadata']['processing']
                task_id = proc.workload_id

                # Client.retryTask(task_id)
                status, out = Client.retryTask(task_id, newParams={})
                self.logger.warn("Retry processing(%s) with task id(%s): %s, %s" % (processing['processing_id'], task_id, status, out))
                # Client.reactivateTask(task_id)
                # Client.resumeTask(task_id)
        except Exception as ex:
            msg = "Failed to check the processing (%s) status: %s" % (str(processing['processing_id']), str(ex))
            raise exceptions.IDDSException(msg)

    def poll_processing_updates(self, processing, input_output_maps):
        """
        *** Function called by Carrier agent.
        """
        updated_contents = []
        update_processing = {}
        reset_expired_at = False
        reactive_contents = []
        # self.logger.debug("poll_processing_updates, input_output_maps: %s" % str(input_output_maps))

        if processing:
            proc = processing['processing_metadata']['processing']
            if proc.tocancel:
                self.logger.info("Cancelling processing (processing id: %s, jediTaskId: %s)" % (processing['processing_id'], proc.workload_id))
                self.kill_processing_force(processing)
                # self.kill_processing(processing)
                proc.tocancel = False
                proc.polling_retries = 0
            elif proc.tosuspend:
                self.logger.info("Suspending processing (processing id: %s, jediTaskId: %s)" % (processing['processing_id'], proc.workload_id))
                self.kill_processing_force(processing)
                # self.kill_processing(processing)
                proc.tosuspend = False
                proc.polling_retries = 0
            elif proc.toresume:
                self.logger.info("Resuming processing (processing id: %s, jediTaskId: %s)" % (processing['processing_id'], proc.workload_id))
                self.reactivate_processing(processing)
                reset_expired_at = True
                proc.toresume = False
                proc.polling_retries = 0
                proc.has_new_updates()
                # reactive_contents = self.reactive_contents(input_output_maps)
            # elif self.is_processing_expired(processing):
            elif proc.toexpire:
                self.logger.info("Expiring processing (processing id: %s, jediTaskId: %s)" % (processing['processing_id'], proc.workload_id))
                self.kill_processing(processing)
                proc.toexpire = False
                proc.polling_retries = 0
            elif proc.tofinish or proc.toforcefinish:
                self.logger.info("Finishing processing (processing id: %s, jediTaskId: %s)" % (processing['processing_id'], proc.workload_id))
                self.kill_processing(processing)
                proc.tofinish = False
                proc.toforcefinish = False
                proc.polling_retries = 0

            processing_status, poll_updated_contents, new_input_output_maps = self.poll_panda_task(processing=processing, input_output_maps=input_output_maps)
            self.logger.debug("poll_processing_updates, processing_status: %s" % str(processing_status))
            self.logger.debug("poll_processing_updates, update_contents: %s" % str(poll_updated_contents))

            if poll_updated_contents:
                proc.has_new_updates()
            for content in poll_updated_contents:
                updated_content = {'content_id': content['content_id'],
                                   'substatus': content['substatus'],
                                   'content_metadata': content['content_metadata']}
                updated_contents.append(updated_content)

            content_substatus = {'finished': 0, 'unfinished': 0}
            for map_id in input_output_maps:
                outputs = input_output_maps[map_id]['outputs']
                for content in outputs:
                    if content.get('substatus', ContentStatus.New) != ContentStatus.Available:
                        content_substatus['unfinished'] += 1
                    else:
                        content_substatus['finished'] += 1

            if processing_status in [ProcessingStatus.SubFinished, ProcessingStatus.Finished, ProcessingStatus.Failed] and updated_contents:
                self.logger.info("Processing %s is terminated, but there are still contents to be flushed. Waiting." % (proc.workload_id))
                # there are still polling contents, should not terminate the task.
                processing_status = ProcessingStatus.Running

            if processing_status in [ProcessingStatus.SubFinished] and content_substatus['finished'] > 0 and content_substatus['unfinished'] == 0:
                # found that a 'done' panda task has got a 'finished' status. Maybe in this case 'finished' is a transparent status.
                if proc.polling_retries is None:
                    proc.polling_retries = 0

            if processing_status in [ProcessingStatus.SubFinished, ProcessingStatus.Finished, ProcessingStatus.Failed]:
                if proc.polling_retries is not None and proc.polling_retries < 3:
                    self.logger.info("processing %s polling_retries(%s) < 3, keep running" % (processing['processing_id'], proc.polling_retries))
                    processing_status = ProcessingStatus.Running
                    proc.polling_retries += 1
            else:
                proc.polling_retries = 0

            if proc.in_operation_time():
                processing_status = ProcessingStatus.Running

            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': processing_status}}
            if reset_expired_at:
                processing['expired_at'] = None
                update_processing['parameters']['expired_at'] = None
                proc.polling_retries = 0
                # if (processing_status in [ProcessingStatus.SubFinished, ProcessingStatus.Finished, ProcessingStatus.Failed]
                #     or processing['status'] in [ProcessingStatus.Resuming]):   # noqa W503
                # using polling_retries to poll it again when panda may update the status in a delay(when issuing retryTask, panda will not update it without any delay).
                update_processing['parameters']['status'] = ProcessingStatus.Resuming
            proc.status = update_processing['parameters']['status']

        self.logger.debug("poll_processing_updates, task: %s, update_processing: %s" %
                          (proc.workload_id, str(update_processing)))
        self.logger.debug("poll_processing_updates, task: %s, updated_contents: %s" %
                          (proc.workload_id, str(updated_contents)))
        self.logger.debug("poll_processing_updates, task: %s, reactive_contents: %s" %
                          (proc.workload_id, str(reactive_contents)))
        return update_processing, updated_contents + reactive_contents, new_input_output_maps

    def get_status_statistics(self, registered_input_output_maps):
        status_statistics = {}
        for map_id in registered_input_output_maps:
            outputs = registered_input_output_maps[map_id]['outputs']

            for content in outputs:
                if content['status'].name not in status_statistics:
                    status_statistics[content['status'].name] = 0
                status_statistics[content['status'].name] += 1
        self.status_statistics = status_statistics
        self.logger.debug("registered_input_output_maps, status_statistics: %s" % str(status_statistics))
        return status_statistics

    def syn_work_status(self, registered_input_output_maps, all_updates_flushed=True, output_statistics={}, to_release_input_contents=[]):
        super(ATLASPandaWork, self).syn_work_status(registered_input_output_maps, all_updates_flushed, output_statistics, to_release_input_contents)
        # self.get_status_statistics(registered_input_output_maps)
        self.status_statistics = output_statistics

        self.logger.debug("syn_work_status, self.active_processings: %s" % str(self.active_processings))
        self.logger.debug("syn_work_status, self.has_new_inputs(): %s" % str(self.has_new_inputs))
        self.logger.debug("syn_work_status, coll_metadata_is_open: %s" %
                          str(self.collections[self._primary_input_collection].coll_metadata['is_open']))
        self.logger.debug("syn_work_status, primary_input_collection_status: %s" %
                          str(self.collections[self._primary_input_collection].status))

        self.logger.debug("syn_work_status(%s): is_processings_terminated: %s" % (str(self.get_processing_ids()), str(self.is_processings_terminated())))
        self.logger.debug("syn_work_status(%s): is_input_collections_closed: %s" % (str(self.get_processing_ids()), str(self.is_input_collections_closed())))
        self.logger.debug("syn_work_status(%s): has_new_inputs: %s" % (str(self.get_processing_ids()), str(self.has_new_inputs)))
        self.logger.debug("syn_work_status(%s): has_to_release_inputs: %s" % (str(self.get_processing_ids()), str(self.has_to_release_inputs())))
        self.logger.debug("syn_work_status(%s): to_release_input_contents: %s" % (str(self.get_processing_ids()), str(to_release_input_contents)))

        if self.is_processings_terminated() and self.is_input_collections_closed() and not self.has_new_inputs and not self.has_to_release_inputs() and not to_release_input_contents:
            # if not self.is_all_outputs_flushed(registered_input_output_maps):
            if not all_updates_flushed:
                self.logger.warn("The work processings %s is terminated. but not all outputs are flushed. Wait to flush the outputs then finish the transform" % str(self.get_processing_ids()))
                return

            if self.is_processings_finished():
                self.status = WorkStatus.Finished
            elif self.is_processings_subfinished():
                self.status = WorkStatus.SubFinished
            elif self.is_processings_failed():
                self.status = WorkStatus.Failed
            elif self.is_processings_expired():
                self.status = WorkStatus.Expired
            elif self.is_processings_cancelled():
                self.status = WorkStatus.Cancelled
            elif self.is_processings_suspended():
                self.status = WorkStatus.Suspended
        elif self.is_processings_running():
            self.status = WorkStatus.Running
        else:
            self.status = WorkStatus.Transforming

        if self.is_processings_terminated() or self.is_processings_running() or self.is_processings_started():
            self.started = True
